import pytest

from pysaleryd.skeleton import main

__author__ = "Björn Dalfors"
__copyright__ = "Björn Dalfors"
__license__ = "MIT"


# def test_fib():
#     """API Tests"""
#     assert fib(1) == 1
#     assert fib(2) == 1
#     assert fib(7) == 13
#     with pytest.raises(AssertionError):
#         fib(-10)


def test_main(capsys):
    """CLI Tests"""
    # capsys is a pytest fixture that allows asserts against stdout/stderr
    # https://docs.pytest.org/en/stable/capture.html
    main(["--host", "192.168.1.151", "--port", "3001", "-t", "3"])
    captured = capsys.readouterr()
    assert True



import logging
import asyncio
from .websocket import WSClient, Signal, State
from aiohttp import ClientSession

_LOGGER: logging.Logger = logging.getLogger(__package__)


class Client:
    """Client to manage communication with HRV"""

    def __init__(self, url: str, port: int, session: ClientSession = None):
        self._url = url
        self._port = port
        self._session = ClientSession() if session is None else session
        self._state = State.NONE
        self._data = {}
        self._handlers = []
        self._socket = WSClient(self._session, self._url, self._port, self._handler)

    def add_handler(self, handler):
        """Add event handler"""
        self._handlers.append(handler)

    async def _handler(self, signal: Signal, data: dict = None, state: State = None):
        """Call handlers if data"""
        if signal == Signal.DATA:
            for handler in self._handlers:
                handler(data)

            parsed_message = self._parse_message(data)
            if parsed_message:
                (key, value) = parsed_message
                self._data[key] = value

        elif signal == Signal.CONNECTION_STATE:
            self._state = self._socket.state

    def _parse_message(self, msg):
        """parse socket message"""
        parsed = None

        try:
            if msg[0] == "#":
                if msg[1] == "$":
                    # ack message, strip ack char and treat as state update
                    msg = msg[1::]
                value = msg[1::].split(":")[1].strip()
                if msg[1] != "*":
                    # messages not beginning with * are arrays of integers
                    # [value, min, max] or [value, min, max, time_left]
                    value = [
                        int(v.strip()) if v.strip().isnumeric() else v.strip()
                        for v in value.split("+")
                    ]
                key = msg[1::].split(":")[0]
                parsed = (key, value)
        except Exception:
            _LOGGER.warning("Failed to parse message %s", msg, exc_info=True)
        return parsed

    @property
    def state(self):
        """Get socket state"""
        return self._socket.state

    @property
    def data(self):
        """Get data"""
        return self._data

    async def async_get_data(self):
        """Async get data"""
        return self.data

    async def connect(self):
        """Start socket and wait for connection"""
        self._socket.start()
        while self._socket.state != State.RUNNING:
            await asyncio.sleep(1)

    async def send_command(self, key, value):
        """Send command to HRV"""

        async def ack_command():
            """Should probably ack command here, just sleep for now"""
            await asyncio.sleep(2)

        await self._socket.send_message(f"#{key}:{value}\r")
        await asyncio.gather(ack_command())



"""Python library to connect HRV and Home Assistant to work together."""

from asyncio import create_task, get_running_loop
from collections import deque
from collections.abc import Callable, Coroutine
import enum
import logging
from typing import Any, Final
import datetime
import aiohttp
import asyncio

LOGGER = logging.getLogger(__package__)


class Signal(enum.Enum):
    """What is the content of the callback."""

    CONNECTION_STATE = "state"
    DATA = "data"


class State(enum.Enum):
    """State of the connection."""

    NONE = ""
    RETRYING = "retrying"
    RUNNING = "running"
    STOPPED = "stopped"


RETRY_TIMER: Final = 15
QUEUE_LEN: Final = 100


class WSClient:
    """Websocket transport, session handling, message generation."""

    def __init__(
        self,
        session: aiohttp.ClientSession,
        host: str,
        port: int,
        callback
    ) -> None:
        """Create resources for websocket communication."""
        self.session = session
        self.host = host
        self.port = port
        self.session_handler_callback = callback

        self.loop = get_running_loop()
        self._ws = None
        self._state = self._previous_state = State.NONE

    @property
    def state(self) -> State:
        """State of websocket."""
        return self._state

    def set_state(self, value: State) -> None:
        """Set state of websocket and store previous state."""
        self._previous_state = self._state
        self._state = value

    def state_changed(self) -> None:
        """Signal state change."""
        create_task(self.session_handler_callback(
            Signal.CONNECTION_STATE, state=self._state))

    def start(self) -> None:
        """Start websocket and update its state."""
        create_task(self.running())

    async def running(self) -> None:
        """Start websocket connection."""
        if self._state == State.RUNNING:
            return

        url = f"http://{self.host}:{self.port}"

        try:
            LOGGER.info("Connecting to websocket (%s:%s)", self.host, self.port)
            self._ws = await self.session.ws_connect(url, timeout=10)
            LOGGER.info("Connected to websocket (%s:%s)", self.host, self.port)
            await self._ws.send_str("#\r")
            await self._ws.receive_str()
            self.set_state(State.RUNNING)
            self.state_changed()

            async for msg in self._ws:

                if self._state == State.STOPPED:
                    await self._ws.close()
                    break

                if msg.type == aiohttp.WSMsgType.TEXT:
                    create_task(self.session_handler_callback(
                        Signal.DATA, data=msg.data))
                    LOGGER.debug("%s Received: %s", datetime.datetime.now(), msg.data)
                    continue

                if msg.type == aiohttp.WSMsgType.CLOSED:
                    LOGGER.warning("Connection closed (%s)", self.host)
                    break

                if msg.type == aiohttp.WSMsgType.ERROR:
                    LOGGER.error("Websocket error (%s)", self.host)
                    break

        except aiohttp.ClientConnectorError:
            if self._state != State.RETRYING:
                LOGGER.error("Websocket is not accessible (%s)", self.host)

        except Exception as err:
            if self._state != State.RETRYING:
                LOGGER.error("Unexpected error (%s) %s", self.host, err)

        if self._state != State.STOPPED:
            self.retry()

    def stop(self) -> None:
        """Close websocket connection."""
        self.set_state(State.STOPPED)
        LOGGER.info("Shutting down connection to websocket (%s)", self.host)

    def retry(self) -> None:
        """Retry to connect to websocket.

        Do an immediate retry without timer and without signalling state change.
        Signal state change only after first retry fails.
        """
        if self._state == State.RETRYING and self._previous_state == State.RUNNING:
            LOGGER.info(
                "Reconnecting to websocket (%s) failed, scheduling retry at an interval of %i seconds",
                self.host,
                RETRY_TIMER,
            )
            self.state_changed()

        self.set_state(State.RETRYING)

        if self._previous_state == State.RUNNING:
            LOGGER.info("Reconnecting to websocket (%s)", self.host)
            self.start()
            return

        self.loop.call_later(RETRY_TIMER, self.start)

    async def send_message(self, message):
        """Send message to websocket"""
        if self._state != State.RUNNING:
            raise Exception(
                "Failed to send message {} to websocket. Socket state is {}".format(
                    message,
                    self.state,
                )
            )
        await self._ws.send_str(message)
